/*
 * rs_shelling.c
 * Search shellable h-assignment by reverse search.
 *
 * by Masahiro Hachimori, Aug 2002.
 */

#include <stdio.h>
#include <stdlib.h>
#include "s_complex.h"
#include <time.h>

/*
 *
 *
 *
 */

/*
 * function to compute facet-ridge incidences.
 */
void get_incidence(fadj, radj, numf, numr, faces)
  int **fadj; /* list of incident ridges */
  int **radj; /* list of incident facets */
  int numf;
  int numr;
  Simplices *faces;
{
  Simplices *fhead, *rhead, *tmpf, *tmpr;
  int dim, i, j, num;

  dim = faces->vert[0];
  fhead = faces;
  rhead = faces;
  while (rhead->vert[0] == dim)
    rhead = rhead->next;

  i = 0;
  tmpf = fhead;
  while(tmpf->vert[0] == dim){
    fadj[i] = calloc(sizeof(int), (dim + 2));
    fadj[i][0] = dim + 1; /* the number of incident ridges is d+1 */
    tmpr = rhead;
    j = 0;
    num = 1;
    while(tmpr->vert[0] == dim-1){
      if (face_contained(tmpr->vert, tmpf->vert)){
        fadj[i][num] = j;
        num ++;
        if (num == dim + 2)
          break;
        }
      if (tmpr->next == NULL)
        break;
      else{
        tmpr = tmpr->next;
        j ++;}
      }
    i ++;
    tmpf = tmpf->next;
    }

  i = 0;
  tmpr = rhead;
  while(tmpr->vert[0] == dim-1){
    tmpf = fhead;
    num = 1; /* first just count incident facets */
    while(tmpf->vert[0] == dim){
      if (face_contained(tmpr->vert, tmpf->vert)){
        num ++;
        }
      tmpf = tmpf->next;
      }
    radj[i] = calloc(sizeof(int), num);
    radj[i][0] = num-1;
    num = 1; /* then write incident facets to the list */
    j = 0;
    tmpf = fhead;
    while(tmpf->vert[0] == dim){
      if (face_contained(tmpr->vert, tmpf->vert)){
        radj[i][num] = j;
        num ++;
        }
      tmpf = tmpf->next;
      j ++;
      }
    i ++;
    if (tmpr->next == NULL)
      break;
    else 
      tmpr = tmpr->next;
    }

  return;
}


/*
 * increment h-vector
 */
void increment_hvec(i, b_arr, m_arr)
  int i;
  int b_arr[MAXDIM][MAXDIM];
  int m_arr[MAXDIM][MAXDIM];
{
  int j, k;

  for (j = 1; j <= i-1; j ++) {
    if (b_arr[i][j] < b_arr[i][j+1]) {
      b_arr[i][j] ++;
      if (b_arr[i][j] < 1)
        m_arr[i][j] = 0;
      else if (b_arr[i][j] == 1)
        m_arr[i][j] = 1;
      else 
        m_arr[i][j] = m_arr[i][j] * (b_arr[i][j] + j - 2) / (b_arr[i][j] - 1);
      for (k = 1; k <= j-1; k ++){
        b_arr[i][k] = 0;
        m_arr[i][k] = 0;
        }
      return;
      }
    }
  /* if all entry are equall */
  b_arr[i][i] ++;
  if (b_arr[i][i] < 1)
    m_arr[i][i] = 0;
  else if (b_arr[i][i] == 1)
    m_arr[i][i] = 1;
  else 
    m_arr[i][i] = m_arr[i][i] * (b_arr[i][i] + i - 2) / (b_arr[i][i] - 1);
  for (j = 1; j <= i-1; j ++){
    b_arr[i][j] = 0;
    m_arr[i][j] = 0;
    }

  return;

}

/*
 * decrement h-vector
 */
void decrement_hvec(i, b_arr, m_arr)
  int i;
  int b_arr[MAXDIM][MAXDIM];
  int m_arr[MAXDIM][MAXDIM];
{
  int j, k;

  for (j = 1; j <= i; j ++){
    if ((j == 1 && b_arr[i][j] > 0) ||
        (j > 1 && b_arr[i][j-1] < b_arr[i][j])) {
      b_arr[i][j] --;
      if (b_arr[i][j] < 1)
        m_arr[i][j] = 0;
      else if (b_arr[i][j] == 1)
        m_arr[i][j] = 1;
      else
        m_arr[i][j] = m_arr[i][j] * (b_arr[i][j] ) / (b_arr[i][j] + j - 1);
      if (j > 1 && b_arr[i][j-1] == 0)
        for (k = j-1; k >= 1; k --){
          b_arr[i][k] = b_arr[i][j]; 
          if (b_arr[i][k] < 1)
            m_arr[i][k] = 0;
          else if (b_arr[i][k] == 1)
            m_arr[i][k] = 1;
          else 
            m_arr[i][k] = m_arr[i][k+1] * k / (b_arr[i][k]+k-1);
          }
      return;
      }
    }

}

/*
 * function of shellability checking.
 */
int rs_shellable(facets)
  Simplices *facets;
{
  int f_vec[MAXDIM], h_vec[MAXDIM], tmph_vec[MAXDIM];
  Simplices *faces, *tmpfaces;
  int dim, numf, numr, num_assigned;
  int **fadj, **radj, **fori, **rori, *hassign;
  int i, j, k, l, m, idx, b;
  int *fcancel, maxcancel, tmpmax, *rval, *fbd;
  int b_arr[MAXDIM][MAXDIM], m_arr[MAXDIM][MAXDIM];
  int tmpi[MAXDIM], dec_flag;
unsigned long nn = 0;
int sum;

  /* preparation of data from facet list*/
  dim = facets->vert[0];
  if (dim == 0) return 1;

  faces = facets2faces(facets);

  get_fvector(faces, f_vec);
  f_to_h(f_vec, h_vec, facets->vert[0]);
  /* check nonnegativity of h-vector */
  for (i = 0; i <= dim + 1; i ++)
    if (h_vec[i] < 0) {printf("h-vector negative\n"); return 0;}
  /* copy h-vector to temporary vector used in the loop */
  for (i = 0; i <= dim + 1; i ++)
    tmph_vec[i] = h_vec[i];
  /* get binom expansion and combination numbers */
  /* (b_arr ... b_i - 1,  m_arr ... \choose{b_i - 1}{j - 1},
      where a_i: binomial expansioin, b_i = a_i - j.) */
  /* (we need to check M-ness from 2 to dim+1) */
  for (i = 1; i <= dim + 1; i ++) {
    binom_expand(h_vec[i], i, b_arr[i]);
    }
  for (i = 1; i <= dim + 1; i ++) {
    for (j = 1; j <= i; j ++) {
      b_arr[i][j] = b_arr[i][j] - j + 1;
      m_arr[i][j] = binom(b_arr[i][j] + j - 2, j - 1);
      }
    }
  for (i = 1; i <= dim + 1; i ++) {
    if (h_vec[i] > 0) decrement_hvec(i, b_arr, m_arr);
  }
/* test */
#ifdef TESTHVEC
#define I 3
for (i = 0; i < 30; i ++){
for (j = 1; j <= I; j ++)
  printf("%d ", b_arr[I][j]);
printf("  ");
for (j = 1; j <= I; j ++)
  printf("%d ", b_arr[I][j]+j-2);
printf("  ");
for (j = 1; j <= I; j ++)
  printf("%d ", m_arr[I][j]);
printf("\n");
increment_hvec(I, b_arr, m_arr);
}
printf("---\n");
for (i = 0; i < 30; i ++){
for (j = 1; j <= I; j ++)
  printf("%d ", b_arr[I][j]);
printf("  ");
for (j = 1; j <= I; j ++)
  printf("%d ", b_arr[I][j]+j-2);
printf("  ");
for (j = 1; j <= I; j ++)
  printf("%d ", m_arr[I][j]);
printf("\n");
decrement_hvec(I, b_arr, m_arr);
}
exit(0);
#endif
/********/
  /* M-sequence check */
  if (h_vec[0] != 1) {printf("not M-sequence (h_0 not 1).\n"); return 0;}
  for (i = 2; i <= dim+1; i ++) {
    sum = 0;
    for (j = 1; j <= i; j ++) {
      sum += m_arr[i][j];
      }
    if (sum > h_vec[i-1]) {printf("not M-sequence.(h_{%d-1} smaller)\n", i);
                           return 0;}
    }

  tmpfaces = faces;
  numf = 0; /* count the number of facets */
  while(tmpfaces->vert[0] == dim){
    numf ++;
    tmpfaces = tmpfaces->next; }
  numr = 0;
  while(tmpfaces->vert[0] == dim-1){
    numr++; /* count the number of ridges */
    if (tmpfaces->next != NULL)
      tmpfaces = tmpfaces->next;
    else
      break;
    }
  fadj = calloc(sizeof(int*), numf);
  radj = calloc(sizeof(int*), numr);
  printf("%d %d\n", numf, numr);

  get_incidence(fadj, radj, numf, numr, faces);

  rval = calloc(sizeof(int), numr); /* valency of ridges */
  for (i = 0; i < numr; i ++)
    rval[i] = radj[i][0];

  fbd = calloc(sizeof(int), numf); /* dim+1 - # of boundary ridges */
  for (i = 0; i < numf; i ++)
    for (j = 1; j <= fadj[i][0]; j ++)
      if (rval[fadj[i][j]] != 1)
        fbd[i] ++;
    
  hassign = calloc(sizeof(int), numf);
  for (i = 0; i < numf; i ++) hassign[i] = -1; /* -1 means not assigned.*/

  fcancel = calloc(sizeof(int), numf); /* to remember cancellable facets */
  maxcancel = -1;


  /* Now start searching. */

  /* main loop */
  idx = 0; /* starting index of facet */
  num_assigned = 0;
  while(1){
    for (i = 1; i <= dim+1; i ++) tmpi[i] = 0;
    /* look for removable facets */
    for (i = idx; i < numf; i ++){
/* printf("i=%d (num_facets:%d %d)\n", i, numf, num_assigned); */
      if (hassign[i] == -1 && tmph_vec[fbd[i]] > 0){
        tmph_vec[fbd[i]] --;

        /* if the facet is the last one ... */
        if (num_assigned == numf-1) {
          hassign[i] = fbd[i];
          for(i=0; i < numf; i ++)
            if (hassign[i] == -1) printf("*"); else printf("%d", hassign[i]);
          printf("\n");
printf("%lu\n", nn);
          return(1);
          }

        /* 0 should be assigned only in the last.
           if 0-assigning is possible, stop further searching. */
        if(tmph_vec[0] == 0){ 
/* printf("disconnected\n"); */
          tmph_vec[fbd[i]] ++;
          i = numf;
          break;}

        /* M-sequence check */
        dec_flag = 0;
        if (tmpi[fbd[i]] == -1) { /* if already know this removing is bad. */
          tmph_vec[fbd[i]] ++;
          goto nextloop; }
        else if (tmpi[fbd[i]] == 1) { /* if already know the removing is good.*/
          /*decrement_hvec(fbd[i], b_arr, m_arr)*/;}
        else {
          if(tmph_vec[fbd[i]] >= 0) decrement_hvec(fbd[i], b_arr, m_arr);
          dec_flag = 1;
          for (j = 2; j <= dim+1; j ++) {
            sum = 0;
            for (k = 1; k <= j; k ++) {
              sum += m_arr[j][k];
              }
            if (sum > tmph_vec[j-1]) {
/* printf("worked(i = %d)\n", i); */
              if (tmph_vec[fbd[i]] >= 0) increment_hvec(fbd[i], b_arr, m_arr);
              tmph_vec[fbd[i]] ++;
              tmpi[fbd[i]] = -1;
              goto nextloop;  /* (goto next step of the for-loop of i.) */
              }
            }
          /* increment_hvec(fbd[i], b_arr, m_arr); */
          tmpi[fbd[i]] = 1;
          }
       
        idx = i; /* now idx shows the removable facet. */

        tmpmax = maxcancel;
        if (tmpmax > idx) {
          b = 0; /* first, check cancellability of tmpmax */
          for (j = 1; j <= fadj[tmpmax][0]; j ++) {
            if (rval[fadj[tmpmax][j]] == 1)
              for (k = 1; k <= radj[fadj[tmpmax][j]][0]; k ++) 
                if (radj[fadj[tmpmax][j]][k] == idx) {
                  b = 1;
                  break;}
              if (b == 1) break; }
          if (b == 1) { /* if not cancellable, check between tmpmax and idx */
            for (j = tmpmax-1; j > idx; j --) {
              if (fcancel[j] == 1) {
                b = 2;
                for (k = 1; k <= fadj[j][0]; k ++) { 
                  if (rval[fadj[j][k]] == 1)
                    for (l = 1; l <= radj[fadj[j][k]][0]; l ++)
                      if (radj[fadj[j][k]][l] == idx) {
                        b = 3;
                        break; }
                  if (b == 3) break;
                  }
                if (b == 2){
                  tmpmax = j;
                  break;
                  }
                }
              }
            if (j == idx) {
              tmpmax = j;
              }
            }
          }

        /* check whether the removing of the facet is approval or not. */
        if (tmpmax > idx) { /* if larger facet is cancellable, not remove.*/
          tmph_vec[fbd[i]] ++;
          if (dec_flag == 1) 
            if (tmph_vec[fbd[i]] > 0) increment_hvec(fbd[i], b_arr, m_arr);
          continue; }
        else{
          /* renew cancellability */
          b = 0;
          for (j = 1; j <= fadj[i][0]; j ++)
            if (rval[fadj[i][j]] == 1)
              for (k = 1; k <= radj[fadj[i][j]][0]; k ++)
                if (fcancel[radj[fadj[i][j]][k]] >= 1){
                  fcancel[radj[fadj[i][j]][k]] ++; /* adjacent cancellable
                                                      becomes uncancellable.*/
                  if (radj[fadj[i][j]][k] == maxcancel)
                    b = 1;
                  }
          fcancel[i] = 1; /* removed one is cancellable */
          if (i > maxcancel) maxcancel = i;
          else
            if (b == 1)
              for (j = maxcancel-1; j >= i; j --)
                if (fcancel[j] == 1){
                  maxcancel = j;
                  break;}
          /* removing step */
          num_assigned ++;
          hassign[i] = fbd[i];
          for (j = 1; j <= fadj[i][0]; j ++) {
            rval[fadj[i][j]] --;
            if (rval[fadj[i][j]] == 1)
              for (k = 1; k <= radj[fadj[i][j]][0]; k ++)
                if (radj[fadj[i][j]][k] != i &&
                    hassign[radj[fadj[i][j]][k]] == -1) {
                  fbd[radj[fadj[i][j]][k]] --;
                  break;}
            }
          if (dec_flag == 0) 
            if (tmph_vec[fbd[i]] >= 0) decrement_hvec(hassign[i], b_arr, m_arr);

          idx = 0;
          break; 
          }
        }
      nextloop:
      ;
      }

    if (i == numf) { /* if no removable facets are found ... */
      if (num_assigned == 0){
printf("%lu\n", nn);
        return(0);  /* if in the root, conclude no shellable h-assignment.*/
        }
      else {          /* otherwise go one step backwards. */
        j = maxcancel;
/* if (j == -1) {printf("error!\n");exit (-1);} */
              for (k = 1; k <= fadj[j][0]; k ++) /* resume cancellability */
                if (rval[fadj[j][k]] == 0)
                  for (l = 1; l <= radj[fadj[j][k]][0]; l ++)
                    if (fcancel[radj[fadj[j][k]][l]] >= 2){
                      fcancel[radj[fadj[j][k]][l]] --;
                      if (fcancel[radj[fadj[j][k]][l]] == 1 &&
                          maxcancel < radj[fadj[j][k]][l])
                        maxcancel = radj[fadj[j][k]][l]; }
              fcancel[j] = 0;
              if (maxcancel == j){
                for (k = j-1; k >= 0; k --)
                  if (fcancel[k] == 1){
                    maxcancel = k;
                    break;}
                if (k == -1) maxcancel = -1;}
              /* unremoving step */
              tmph_vec[hassign[j]] ++;
              if (tmph_vec[hassign[j]] > 0) increment_hvec(hassign[j], b_arr, m_arr);
              hassign[j] = -1;
              num_assigned --;
              for (k = 1; k <= fadj[j][0]; k ++) {
                rval[fadj[j][k]] ++;
                if (rval[fadj[j][k]] == 2)
                  for (l = 1; l <= radj[fadj[j][k]][0]; l ++)
                  if (radj[fadj[j][k]][l] != j &&
                      hassign[radj[fadj[j][k]][l]] == -1) {
                    fbd[radj[fadj[j][k]][l]] ++;
                    break;}
                }
              /* --- */
              idx = j + 1;
nn++;
        /* for debugging only */
/*      if (j == -1) {
          printf("error!\n");
          exit(-1);
          } */
        }
      }
    }

}

